import csv

from bonobo.config import Option, use_raw_input, use_context
from bonobo.config.options import Method, RenamedOption
from bonobo.constants import NOT_MODIFIED
from bonobo.nodes.io.base import FileHandler
from bonobo.nodes.io.file import FileReader, FileWriter
from bonobo.util import ensure_tuple
from bonobo.util.bags import BagType


class CsvHandler(FileHandler):
    """

    .. attribute:: delimiter

        The CSV delimiter.

    .. attribute:: quotechar

        The CSV quote character.

    .. attribute:: fields

        The list of column names, if the CSV does not contain it as its first line.

    """

    # Dialect related options
    delimiter = Option(str, default=csv.excel.delimiter, required=False)
    quotechar = Option(str, default=csv.excel.quotechar, required=False)
    escapechar = Option(str, default=csv.excel.escapechar, required=False)
    doublequote = Option(str, default=csv.excel.doublequote, required=False)
    skipinitialspace = Option(str, default=csv.excel.skipinitialspace, required=False)
    lineterminator = Option(str, default=csv.excel.lineterminator, required=False)
    quoting = Option(str, default=csv.excel.quoting, required=False)

    # Fields (renamed from headers)
    headers = RenamedOption('fields')
    fields = Option(ensure_tuple, required=False)

    def get_dialect_kwargs(self):
        return {
            'delimiter': self.delimiter,
            'quotechar': self.quotechar,
            'escapechar': self.escapechar,
            'doublequote': self.doublequote,
            'skipinitialspace': self.skipinitialspace,
            'lineterminator': self.lineterminator,
            'quoting': self.quoting,
        }


@use_context
class CsvReader(FileReader, CsvHandler):
    """
    Reads a CSV and yield the values as dicts.
    """

    skip = Option(int, default=0, __doc__='''
        If set and greater than zero, the reader will skip this amount of lines.
    ''')

    @Method(
        positional=False,
        __doc__='''
            Builds the CSV reader, a.k.a an object we can iterate, each iteration giving one line of fields, as an
            iterable.
            
            Defaults to builtin csv.reader(...), but can be overriden to fit your special needs.
        '''
    )
    def reader_factory(self, file):
        return csv.reader(file, **self.get_dialect_kwargs())

    def read(self, file, context, *, fs):
        context.setdefault('skipped', 0)
        reader = self.reader_factory(file)
        skip = self.skip

        if not context.output_type:
            context.set_output_fields(self.fields or next(reader))

        for row in reader:
            if context.skipped < skip:
                context.skipped += 1
                continue
            yield tuple(row)

    __call__ = read


@use_context
class CsvWriter(FileWriter, CsvHandler):
    @Method(
        __doc__='''
            Builds the CSV writer, a.k.a an object we can pass a field collection to be written as one line in the
            target file.
            
            Defaults to builtin csv.writer(...).writerow, but can be overriden to fit your special needs.
        '''
    )
    def writer_factory(self, file):
        return csv.writer(file, **self.get_dialect_kwargs()).writerow

    def write(self, file, context, *values, fs):
        context.setdefault('lineno', 0)
        fields = context.get_input_fields()

        if not context.lineno:
            context.writer = self.writer_factory(file)

            if fields:
                context.writer(fields)
                context.lineno += 1

        if fields:
            if len(values) != len(fields):
                raise ValueError(
                    'Values length differs from input fields length. Expected: {}. Got: {}. Values: {!r}.'.format(
                        len(fields), len(values), values
                    )
                )
            context.writer(values)
        else:
            for arg in values:
                context.writer(ensure_tuple(arg))

        return NOT_MODIFIED

    __call__ = write



"""
This package is considered private, and should only be used within bonobo.

"""

import json
import os
import runpy

import bonobo
from bonobo.util import cast


class _RequiredModule:
    def __init__(self, dct):
        self.__dict__ = dct


class _ModulesRegistry(dict):
    @property
    def pathname(self):
        return os.getcwd()

    def require(self, name):
        if name not in self:
            bits = name.split('.')
            filename = os.path.join(self.pathname, *bits[:-1], bits[-1] + '.py')
            self[name] = _RequiredModule(runpy.run_path(filename, run_name=name))
        return self[name]


def _parse_option(option):
    """
    Parse a 'key=val' option string into a python (key, val) pair

    :param option: str
    :return: tuple
    """
    try:
        key, val = option.split('=', 1)
    except ValueError:
        return option, True

    try:
        val = json.loads(val)
    except json.JSONDecodeError:
        pass

    return key, val


def _resolve_options(options=None):
    """
    Resolve a collection of option strings (eventually coming from command line) into a python dictionary.

    :param options: tuple[str]
    :return: dict
    """
    if options:
        return dict(map(_parse_option, options))
    return dict()


@cast(tuple)
def _resolve_transformations(transformations):
    """
    Resolve a collection of strings into the matching python objects, defaulting to bonobo namespace if no package is provided.

    Syntax for each string is path.to.package:attribute

    :param transformations: tuple(str)
    :return: tuple(object)
    """
    registry = _ModulesRegistry()
    transformations = transformations or []
    for t in transformations:
        try:
            mod, attr = t.split(':', 1)
            yield getattr(registry.require(mod), attr)
        except ValueError:
            yield getattr(bonobo, t)



import bonobo


def get_argument_parser(parser=None):
    parser = bonobo.get_argument_parser(parser=parser)

    parser.add_argument(
        '--limit',
        '-l',
        type=int,
        default=None,
        help='If set, limits the number of processed lines.'
    )
    parser.add_argument(
        '--print',
        '-p',
        action='store_true',
        default=False,
        help='If set, pretty prints before writing to output file.'
    )

    return parser


def get_graph_options(options):
    _limit = options.pop('limit', None)
    _print = options.pop('print', False)

    return {
        '_limit': (bonobo.Limit(_limit), ) if _limit else (),
        '_print': (bonobo.PrettyPrinter(), ) if _print else (),
    }



"""
This transformation extracts the environment and prints it, sorted alphabetically, one item per line.

Used in the bonobo tests around environment management.

"""
import os

import bonobo


def extract_environ():
    """Yield all the system environment."""
    yield from sorted(os.environ.items())


def get_graph():
    graph = bonobo.Graph()
    graph.add_chain(extract_environ, print)

    return graph


if __name__ == '__main__':
    parser = bonobo.get_argument_parser()
    with bonobo.parse_args(parser):
        bonobo.run(get_graph())



import bonobo
from bonobo.commands import BaseGraphCommand


class InspectCommand(BaseGraphCommand):
    handler = staticmethod(bonobo.inspect)

    def add_arguments(self, parser):
        super(InspectCommand, self).add_arguments(parser)
        parser.add_argument('--graph', '-g', dest='format', action='store_const', const='graph')

    def parse_options(self, **options):
        if not options.get('format'):
            raise RuntimeError('You must provide a format (try --graph).')
        return options



import logging
import sys
from contextlib import contextmanager
from logging import ERROR

from mondrian import term

from bonobo.util import deprecated
from bonobo.util.objects import Wrapper, get_name


@contextmanager
def recoverable(error_handler):
    try:
        yield
    except Exception as exc:  # pylint: disable=broad-except
        error_handler(*sys.exc_info(), level=ERROR)


@contextmanager
def unrecoverable(error_handler):
    try:
        yield
    except Exception as exc:  # pylint: disable=broad-except
        error_handler(*sys.exc_info(), level=ERROR)
        raise  # raise unrecoverableerror from x ?


class Lifecycle:
    def __init__(self):
        self._started = False
        self._stopped = False
        self._killed = False
        self._defunct = False

    @property
    def started(self):
        return self._started

    @property
    def stopped(self):
        return self._stopped

    @property
    def killed(self):
        return self._killed

    @property
    def defunct(self):
        return self._defunct

    @property
    def alive(self):
        return self._started and not self._stopped

    @property
    def should_loop(self):
        # TODO XXX started/stopped?
        return not any((self.defunct, self.killed))

    @property
    def status(self):
        """One character status for this node. """
        if self._defunct:
            return '!'
        if not self.started:
            return ' '
        if not self.stopped:
            return '+'
        return '-'

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.stop()

    def get_flags_as_string(self):
        if self._defunct:
            return term.red('[defunct]')
        if self.killed:
            return term.lightred('[killed]')
        if self.stopped:
            return term.lightblack('[done]')
        return ''

    def start(self):
        if self.started:
            raise RuntimeError('This context is already started ({}).'.format(get_name(self)))

        self._started = True

    def stop(self):
        if not self.started:
            raise RuntimeError('This context cannot be stopped as it never started ({}).'.format(get_name(self)))

        self._stopped = True

        if self._stopped:  # Stopping twice has no effect
            return

    def kill(self):
        if not self.started:
            raise RuntimeError('Cannot kill an unstarted context.')

        if self.stopped:
            raise RuntimeError('Cannot kill a stopped context.')

        self._killed = True

    @deprecated
    def handle_error(self, exctype, exc, tb, *, level=logging.ERROR):
        return self.error((exctype, exc, tb), level=level)

    def error(self, exc_info, *, level=logging.ERROR):
        logging.getLogger(__name__).log(level, repr(self), exc_info=exc_info)

    def fatal(self, exc_info, *, level=logging.CRITICAL):
        logging.getLogger(__name__).log(level, repr(self), exc_info=exc_info)
        self._defunct = True

    def as_dict(self):
        return {
            'status': self.status,
            'name': self.name,
            'stats': self.get_statistics_as_string(),
            'flags': self.get_flags_as_string(),
        }


class BaseContext(Lifecycle, Wrapper):
    def __init__(self, wrapped, *, parent=None):
        Lifecycle.__init__(self)
        Wrapper.__init__(self, wrapped)
        self.parent = parent



import textwrap
import types

from bonobo.util.inspect import istype


class Option:
    """
    An Option is a descriptor for Configurable's parameters.

    .. attribute:: type

        Option type allows to provide a callable used to cast, clean or validate the option value. If not provided, or
        None, the option's value will be the exact value user provided.

        (default: None)

    .. attribute:: required

        If an option is required, an error will be raised if no value is provided (at runtime). If it is not, option
        will have the default value if user does not override it at runtime.

        Ignored if a default is provided, meaning that the option cannot be required.

        (default: True)

    .. attribute:: positional

        If this is true, it'll be possible to provide the option value as a positional argument. Otherwise, it must
        be provided as a keyword argument.

        (default: False)

    .. attribute:: default

        Default value for non-required options.

        (default: None)

    Example:

        .. code-block:: python

            from bonobo.config import Configurable, Option

            class Example(Configurable):
                title = Option(str, required=True, positional=True)
                keyword = Option(str, default='foo')

                def __call__(self, s):
                    return self.title + ': ' + s + ' (' + self.keyword + ')'

            example = Example('hello', keyword='bar')

    """

    _creation_counter = 0

    def __init__(self, type=None, *, required=True, positional=False, default=None, __doc__=None):
        self.name = None
        self.type = type
        self.required = required if default is None else False
        self.positional = positional
        self.default = default

        # Docstring formating
        self.__doc__ = __doc__ or None
        if self.__doc__:
            self.__doc__ = textwrap.dedent(self.__doc__.strip('\n')).strip()
            if default:
                self.__doc__ += '\nDefault: {!r}'.format(default)

        # This hack is necessary for python3.5
        self._creation_counter = Option._creation_counter
        Option._creation_counter += 1

    def __get__(self, inst, type_):
        # XXX If we call this on the type, then either return overriden value or ... ???
        if inst is None:
            return vars(type_).get(self.name, self)

        if not self.name in inst._options_values:
            inst._options_values[self.name] = self.get_default()

        return inst._options_values[self.name]

    def __set__(self, inst, value):
        inst._options_values[self.name] = self.clean(value)

    def __repr__(self):
        return '<{positional}{typename} {type}{name} default={default!r}{required}>'.format(
            typename=type(self).__name__,
            type='({})'.format(self.type) if istype(self.type) else '',
            name=self.name,
            positional='*' if self.positional else '**',
            default=self.default,
            required=' (required)' if self.required else '',
        )

    def clean(self, value):
        return self.type(value) if self.type else value

    def get_default(self):
        return self.default() if callable(self.default) else self.default


class RemovedOption(Option):
    def __init__(self, *args, value, **kwargs):
        kwargs['required'] = False
        super(RemovedOption, self).__init__(*args, **kwargs)
        self.value = value

    def clean(self, value):
        if value != self.value:
            raise ValueError(
                'Removed options cannot change value, {!r} must now be {!r} (and you should remove setting the value explicitely, as it is deprecated and will be removed quite soon.'.
                format(self.name, self.value)
            )
        return self.value

    def get_default(self):
        return self.value


class RenamedOption(Option):
    def __init__(self, target, *, positional=False):
        super(RenamedOption, self).__init__(required=False, positional=False)
        self.target = target

    def __get__(self, instance, owner):
        raise ValueError(
            'Trying to get value from renamed option {}, try getting {} instead.'.format(self.name, self.target)
        )

    def clean(self, value):
        raise ValueError(
            'Trying to set value of renamed option {}, try setting {} instead.'.format(self.name, self.target)
        )


class Method(Option):
    """
    A Method is a special callable-valued option, that can be used in three different ways (but for same purpose).

    * Like a normal option, the value can be provided to the Configurable constructor.

        >>> from bonobo.config import Configurable, Method

        >>> class MethodExample(Configurable):
        ...     handler = Method()

        >>> example1 = MethodExample(handler=str.upper)

    * It can be used by a child class that overrides the Method with a normal method.

        >>> class ChildMethodExample(MethodExample):
        ...     def handler(self, s: str):
        ...         return s.upper()

        >>> example2 = ChildMethodExample()

    * Finally, it also enables the class to be used as a decorator, to generate a subclass providing the Method a value.

        >>> @MethodExample
        ... def OtherChildMethodExample(s):
        ...     return s.upper()

        >>> example3 = OtherChildMethodExample()

    It's possible to pass a default implementation to a Method by calling it, making it suitable to use as a decorator.

        >>> class MethodExampleWithDefault(Configurable):
        ...     @Method()
        ...     def handler(self):
        ...         pass

    """

    def __init__(self, *, default=None, required=True, positional=True, __doc__=None):
        super().__init__(None, required=required, positional=positional, __doc__=__doc__)

        # If a callable is provided as default, then use self as if it was used as a decorator
        if default is not None:
            if not callable(default):
                raise ValueError('Method defaults should be callable, if provided.')
            self(default)

    def __get__(self, inst, type_):
        x = super(Method, self).__get__(inst, type_)
        if inst:
            x = types.MethodType(x, inst)
        return x

    def __set__(self, inst, value):
        if not callable(value):
            raise TypeError(
                'Option {!r} ({}) is expecting a callable value, got {!r} object: {!r}.'.format(
                    self.name,
                    type(self).__name__,
                    type(value).__name__, value
                )
            )
        inst._options_values[self.name] = self.type(value) if self.type else value

    def __call__(self, impl):
        if self.default:
            raise RuntimeError('Can only be used once as a decorator.')
        self.default = impl
        self.required = False
        return self

    def get_default(self):
        return self.default



import bonobo

graph = bonobo.Graph(
    bonobo.FileReader('coffeeshops.txt'),
    print,
)


def get_services():
    return {'fs': bonobo.open_examples_fs('datasets')}


if __name__ == '__main__':
    bonobo.run(graph, services=get_services())



import functools
import warnings


def deprecated_alias(alias, func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated function alias {}, use {} instead.".format(alias, func.__name__),
            category=DeprecationWarning,
            stacklevel=2
        )
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used."""

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn(
            "Call to deprecated function {}.".format(func.__name__), category=DeprecationWarning, stacklevel=2
        )
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func



from functools import partial
from time import sleep

from bonobo.config import create_container
from bonobo.constants import BEGIN, END
from bonobo.execution import events
from bonobo.execution.contexts.node import NodeExecutionContext
from bonobo.execution.contexts.plugin import PluginExecutionContext
from whistle import EventDispatcher


class GraphExecutionContext:
    NodeExecutionContextType = NodeExecutionContext
    PluginExecutionContextType = PluginExecutionContext

    TICK_PERIOD = 0.25

    @property
    def started(self):
        return any(node.started for node in self.nodes)

    @property
    def stopped(self):
        return all(node.started and node.stopped for node in self.nodes)

    @property
    def alive(self):
        return any(node.alive for node in self.nodes)

    def __init__(self, graph, plugins=None, services=None, dispatcher=None):
        self.dispatcher = dispatcher or EventDispatcher()
        self.graph = graph
        self.nodes = [self.create_node_execution_context_for(node) for node in self.graph]
        self.plugins = [self.create_plugin_execution_context_for(plugin) for plugin in plugins or ()]
        self.services = create_container(services)

        # Probably not a good idea to use it unless you really know what you're doing. But you can access the context.
        self.services['__graph_context'] = self

        for i, node_context in enumerate(self):
            outputs = self.graph.outputs_of(i)
            if len(outputs):
                node_context.outputs = [self[j].input for j in outputs]
            node_context.input.on_begin = partial(node_context._send, BEGIN, _control=True)
            node_context.input.on_end = partial(node_context._send, END, _control=True)
            node_context.input.on_finalize = partial(node_context.stop)

    def __getitem__(self, item):
        return self.nodes[item]

    def __len__(self):
        return len(self.nodes)

    def __iter__(self):
        yield from self.nodes

    def create_node_execution_context_for(self, node):
        return self.NodeExecutionContextType(node, parent=self)

    def create_plugin_execution_context_for(self, plugin):
        if isinstance(plugin, type):
            plugin = plugin()
        return self.PluginExecutionContextType(plugin, parent=self)

    def write(self, *messages):
        """Push a list of messages in the inputs of this graph's inputs, matching the output of special node "BEGIN" in
        our graph."""

        for i in self.graph.outputs_of(BEGIN):
            for message in messages:
                self[i].write(message)

    def dispatch(self, name):
        self.dispatcher.dispatch(name, events.ExecutionEvent(self))

    def start(self, starter=None):
        self.register_plugins()
        self.dispatch(events.START)
        self.tick(pause=False)
        for node in self.nodes:
            if starter is None:
                node.start()
            else:
                starter(node)
        self.dispatch(events.STARTED)

    def tick(self, pause=True):
        self.dispatch(events.TICK)
        if pause:
            sleep(self.TICK_PERIOD)

    def kill(self):
        self.dispatch(events.KILL)
        for node_context in self.nodes:
            node_context.kill()
        self.tick()

    def stop(self, stopper=None):
        self.dispatch(events.STOP)
        for node_context in self.nodes:
            if stopper is None:
                node_context.stop()
            else:
                stopper(node_context)
        self.tick(pause=False)
        self.dispatch(events.STOPPED)
        self.unregister_plugins()

    def register_plugins(self):
        for plugin_context in self.plugins:
            plugin_context.register()

    def unregister_plugins(self):
        for plugin_context in self.plugins:
            plugin_context.unregister()



import json

import bonobo


def split_one_to_map(line):
    k, v = line.split(', ', 1)
    return {k: v}


class MyJsonWriter(bonobo.JsonWriter):
    prefix, suffix = '{', '}'

    def write(self, fs, file, lineno, **row):
        return bonobo.FileWriter.write(
            self, fs, file, lineno,
            json.dumps(row)[1:-1]
        )


graph = bonobo.Graph(
    bonobo.FileReader('coffeeshops.txt'),
    split_one_to_map,
    MyJsonWriter('coffeeshops.json', fs='fs.output'),
)


def get_services():
    return {
        'fs': bonobo.open_examples_fs('datasets'),
        'fs.output': bonobo.open_fs(),
    }


if __name__ == '__main__':
    bonobo.run(graph, services=get_services())
